from ..strategy import TradingStrategy
from ..structs import MarketData, TradeRequest, TradeResponse
from ..enums import Side, OrderType
from ..logging import log


class SMAStrategy(TradingStrategy):
    def __init__(self, long: int = 20, short: int = 5, *args, **kwargs) -> None:
        super(SMAStrategy, self).__init__(*args, **kwargs)
        self.trades = []
        self.long = long
        self.short = short
        self.position = None

        self.golden = False

        self.long_average = 0.0
        self.short_average = 0.0

    def onFill(self, res: TradeResponse) -> None:
        if res.side == Side.BUY:
            log.info('d->g:bought %.2f @ %.2f' % (res.volume, res.price))
        else:
            log.info('g->d:sold %.2f @ %.2f' % (res.volume, res.price))

    def onTrade(self, data: MarketData) -> bool:
        self.trades.append(data)
        self.trades = self.trades[-1 * self.long:]
        long_average = self.calc_average(self.long)
        short_average = self.calc_average(self.short)

        if self.position is None:
            # was death, now golden
            if self.long_average > self.short_average and \
               long_average < short_average:
                req = TradeRequest(side=Side.BUY,
                                   volume=1,
                                   instrument=data.instrument,
                                   order_type=OrderType.MARKET,
                                   exchange=data.exchange,
                                   price=data.price,
                                   time=data.time)
                log.info("requesting buy : %s", req)
                self.position = self.request(req)
        else:
            # was golden, now death
            if self.long_average < self.short_average and \
               long_average > short_average:

                req = TradeRequest(side=Side.SELL,
                                   volume=1,
                                   instrument=data.instrument,
                                   order_type=OrderType.MARKET,
                                   exchange=data.exchange,
                                   price=data.price,
                                   time=data.time)
                log.info("requesting sell : %s", req)
                self.request(req)
                self.position = None

        self.long_average = long_average
        self.short_average = short_average

    def calc_average(self, count):
        calc = 0
        trades = self.trades[0 - count:]
        for t in trades:
            calc += t.price
        return calc / count

    def onError(self, e) -> None:
        log.critical(e)

    def onChange(self, data: MarketData) -> None:
        pass

    def onCancel(self, data: MarketData) -> None:
        pass

    def onOpen(self, data: MarketData) -> None:
        pass



from mock import patch, MagicMock
from datetime import datetime


class TestExecution:
    def setup(self):
        pass
        # setup() before each test method

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        pass
        # setup_class() before any methods in this class

    @classmethod
    def teardown_class(cls):
        pass
        # teardown_class() after any methods in this class

    def test_init(self):
        from ..execution import Execution
        from ..config import ExecutionConfig

        ex = MagicMock()

        ec = ExecutionConfig()
        e = Execution(ec, ex, MagicMock())
        assert e

    def test_request(self):
        from ..execution import Execution
        from ..enums import Side, PairType, OrderType, ExchangeType, CurrencyType
        from ..config import ExecutionConfig
        from ..structs import TradeRequest, Instrument, Account

        ex = {ExchangeType.COINBASE: MagicMock()}
        accounts = {CurrencyType.BTC: Account(id='1',
                                              currency=CurrencyType.BTC,
                                              balance=1.0,
                                              exchange=MagicMock(),
                                              value=-1,
                                              asOf=datetime.now()),
                    CurrencyType.USD: Account(id='2',
                                              currency=CurrencyType.USD,
                                              balance=1.0,
                                              exchange=MagicMock(),
                                              value=-1,
                                              asOf=datetime.now())}
        ex[ExchangeType.COINBASE].accounts.return_value = accounts
        ex[ExchangeType.COINBASE].buy.return_value.exchange = ExchangeType.COINBASE
        ec = ExecutionConfig()
        e = Execution(ec, ex, accounts)

        req = TradeRequest(side=Side.BUY,
                           instrument=Instrument(underlying=PairType.BTCUSD),
                           order_type=OrderType.MARKET,
                           exchange=ExchangeType.COINBASE,
                           volume=1.0,
                           price=1.0,
                           time=datetime.now())

        resp = e.request(req)

        req = TradeRequest(side=Side.SELL,
                           instrument=Instrument(underlying=PairType.BTCUSD),
                           order_type=OrderType.MARKET,
                           exchange=ExchangeType.COINBASE,
                           volume=1.0,
                           price=1.0,
                           time=datetime.now())

        resp = e.request(req)

        req = TradeRequest(side=Side.BUY,
                           instrument=Instrument(underlying=PairType.BTCUSD),
                           order_type=OrderType.MARKET,
                           exchange=ExchangeType.COINBASE,
                           volume=1.0,
                           price=1.0,
                           time=datetime.now())

        resp = e.request(req)
        e.cancel(resp)
        e.cancelAll()



from setuptools import setup, find_packages, Extension
from distutils.version import LooseVersion
from codecs import open
from setuptools.command.build_ext import build_ext
import io
import os
import os.path
import os
import re
import sys
import sysconfig
import platform
import subprocess
try:
    from shutil import which
    CPU_COUNT = os.cpu_count()
except ImportError:
    # Python2
    try:
        from backports.shutil_which import which
    except ImportError:
        which = lambda x: x  # just rely on path
    import multiprocessing
    CPU_COUNT = multiprocessing.cpu_count()

pjoin = os.path.join
here = os.path.abspath(os.path.dirname(__file__))
PREFIX = sysconfig.get_config_vars()['prefix']
name = 'aat'


requires = [
    'aiohttp>=3.5.4',
    'aiostream>=0.3.1',
    'ccxt>=1.18.529',
    'cycler>=0.10.0',
    'Jinja2>=2.10',
    'matplotlib>=2.2.2',
    'numpy>=1.14.5',
    'pandas>=0.23.3',
    'perspective-python>=0.1.6,<0.4.0',
    'pyarrow>=0.11.1',
    'scipy>=1.1.0',
    'seaborn>=0.8.1',
    'sphinx>=1.7.5',
    'six>=1.10.0',
    'requests>=2.13.0',
    'tornado>=5.1',
    'traitlets>=4.3.2',
    'ujson>=1.35',
    'uvloop>=0.12.2',
    'websocket-client>=0.40.0',
    'yarl==1.1.0',
]

requires_dev = [
    'flake8>=3.7.8',
    'mock',
    'pybind11>=2.4.0',
    'pytest>=4.3.0',
    'pytest-cov>=2.6.1',
    'Sphinx>=1.8.4',
    'sphinx-markdown-builder>=0.5.2',
] + requires


def get_version(file, name='__version__'):
    path = os.path.realpath(file)
    version_ns = {}
    with io.open(path, encoding="utf8") as f:
        exec(f.read(), {}, version_ns)
    return version_ns[name]

version = get_version(pjoin(here, name, '_version.py'))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

if sys.version_info.major < 3 or sys.version_info.minor < 7:
    raise Exception('Must be python3.7 or above')


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        self.cmake_cmd = which('cmake')
        try:
            out = subprocess.check_output([self.cmake_cmd, '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)',
                                                   out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension_cmake(ext)

    def build_extension_cmake(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cfg = 'Debug' if self.debug else 'Release'

        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + os.path.abspath('aat'),
            '-DCMAKE_BUILD_TYPE=' + cfg,
            '-DCPP_BUILD_TESTS=0',
            '-DAAT_PYTHON_VERSION={}'.format(platform.python_version()),
            '-DPYTHON_EXECUTABLE={}'.format(sys.executable),
            '-DPython_ROOT_DIR={}'.format(PREFIX),
            '-DPython_ROOT={}'.format(PREFIX),
            '-DAAT_CMAKE_MODULE_PATH={folder}'.format(folder=os.path.join(ext.sourcedir, 'cmake')),
        ]

        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(
                cfg.upper(),
                extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2' if os.environ.get('DOCKER', '') else '-j{}'.format(CPU_COUNT)]

        env = os.environ.copy()
        env["PYTHONPATH"] = os.path.pathsep.join((os.path.join(os.path.dirname(os.__file__), 'site-packages'), os.path.dirname(os.__file__)))

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call([self.cmake_cmd, os.path.abspath(ext.sourcedir)] + cmake_args, cwd=self.build_temp, env=env, stderr=subprocess.STDOUT)
        subprocess.check_call([self.cmake_cmd, '--build', '.'] + build_args, cwd=self.build_temp, env=env, stderr=subprocess.STDOUT)
        print()  # Add an empty line for cleaner output

setup(
    name=name,
    version=version,
    description='Algorithmic trading library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/timkpaine/{}'.format(name),
    author='Tim Paine',
    author_email='timothy.k.paine@gmail.com',
    license='Apache 2.0',
    install_requires=requires,
    extras_require={
        'dev': requires_dev,
    },
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    keywords='algorithmic trading cryptocurrencies',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'aat=aat:main',
        ],
    },
    ext_modules=[CMakeExtension('aat')],
    cmdclass=dict(build_ext=CMakeBuild),
)



# for coverage
from ..strategy import *


class TestStrategy:
    def setup(self):
        pass
        # setup() before each test method

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        pass
        # setup_class() before any methods in this class

    @classmethod
    def teardown_class(cls):
        pass
        # teardown_class() after any methods in this class



import os
import os.path
from mock import MagicMock


class TestRisk:
    def setup(self):
        from ..config import RiskConfig
        from ..risk import Risk

        rc = RiskConfig()
        rc.max_risk = 100.0
        rc.max_drawdown = 100.0
        rc.total_funds = 100.0

        self.risk = Risk(rc, MagicMock(), MagicMock())

    def test_config_file(self):

        from ..parser import parse_file_config
        import tempfile

        t = tempfile.NamedTemporaryFile('w', delete=False)
        t.write('''[general]
verbose=1
print=0
TradingType=simulation

[exchange]
exchanges=coinbase
currency_pairs=BTCUSD,ETHUSD,LTCUSD,BCHUSD,ETHBTC,LTCBTC,BCHBTC

[strategy]
strategies =
    aat.strategies.buy_and_hold.BuyAndHoldStrategy

[risk]
max_drawdown = 100.0
max_risk = 100.0
total_funds = 10.0
''')
        t.close()

        print(t.name)
        x = parse_file_config(t.name)
        assert x

        os.remove(t.name)



from abc import ABCMeta, abstractmethod
from .callback import Callback
from .enums import Side
from .structs import TradeRequest, TradeResponse


class Strategy(metaclass=ABCMeta):
    '''Strategy interface'''

    def __init__(self, query=None, exchanges=None, *args, **kwargs) -> None:
        self.query = query
        self.exchanges = exchanges

    def setEngine(self, engine) -> None:
        self._te = engine

    @abstractmethod
    def request(self, req: TradeRequest):
        '''request'''


class TradingStrategy(Strategy, Callback):
    def request(self, req: TradeRequest) -> None:
        '''attempt to buy/sell'''
        return self._te.request(req, self)

    def cancel(self, resp: TradeResponse) -> None:
        '''cancel order'''
        return self._te.cancel(resp, self)

    def cancelAll(self):
        '''cancel all orders'''
        return self._te.cancelAll(self)

    def slippage(self, data: TradeResponse) -> TradeResponse:
        '''slippage model. default is pass through'''
        return data

    def transactionCost(self, data: TradeResponse) -> TradeResponse:
        '''txns cost model. default is pass through'''
        return data

    def to_dict(self, *args):
        return {'name': self.__class__.__name__}

    def onAnalyze(self, engine):
        '''onAnalyze'''
        pass
        if not engine:
            return
        import pandas
        import numpy as np
        import matplotlib
        import matplotlib.pyplot as plt
        import seaborn as sns
        matplotlib.rc('font', **{'size': 5})

        # extract data from trading engine
        positions_value = engine.query.positions_value
        portfolio_value = engine.query.portfolio_value

        requests = engine.query.query_tradereqs()
        responses = engine.query.query_traderesps()
        trades = pandas.DataFrame([{'time': x.time, 'price': x.price} for x in engine.query.query_trades(instrument=requests[0].instrument, page=None)])
        trades.set_index(['time'], inplace=True)

        # format into pandas
        pd = pandas.DataFrame(positions_value, columns=['time', 'unrealized', 'realized', 'pnl'])
        pd2 = pandas.DataFrame(portfolio_value, columns=['time', 'value'])
        import ipdb
        ipdb.set_trace()
        pd.set_index(['time'], inplace=True)
        pd2.set_index(['time'], inplace=True)

        # setup charting
        sns.set_style('darkgrid')
        fig = plt.figure(figsize=(13, 7))
        ax1 = fig.add_subplot(411)
        ax2 = fig.add_subplot(412)
        ax3 = fig.add_subplot(413)
        ax4 = fig.add_subplot(414)

        # plot algo performance
        pd2.plot(ax=ax1, y=['value'], legend=False, fontsize=5, rot=0)

        # plot up/down chart
        pd['pos'] = pd['pnl']
        pd['neg'] = pd['pnl']
        pd['pos'][pd['pos'] <= 0] = np.nan
        pd['neg'][pd['neg'] > 0] = np.nan
        pd.plot(ax=ax2, y=['pos', 'neg'], kind='area', stacked=False, color=['green', 'red'], legend=False, linewidth=0, fontsize=5, rot=0)
        ax2.set_ylabel('PnL')

        pd.plot(ax=ax3, y=['unrealized', 'realized', 'pnl'], kind='line', legend=False, fontsize=5, rot=0)
        ax3.legend(loc="upper left")

        # annotate with key data
        ax1.set_title('Performance')
        ax1.set_ylabel('Portfolio value($)')
        for xy in [portfolio_value[0][:2]] + [portfolio_value[-1][:2]]:
            ax1.annotate('$%s' % xy[1], xy=xy, textcoords='data')
            ax4.annotate('$%s' % xy[1], xy=xy, textcoords='data')

        # plot trade intent/trade action
        ax4.set_ylabel('Intent/Action')
        ax4.set_xlabel('Date')

        ax4.plot(trades)
        ax4.plot([x.time for x in requests if x.side == Side.BUY],
                 [x.price for x in requests if x.side == Side.BUY],
                 '2', color='y')
        ax4.plot([x.time for x in requests if x.side == Side.SELL],
                 [x.price for x in requests if x.side == Side.SELL],
                 '1', color='y')
        ax4.plot([x.time for x in responses if x.side == Side.BUY],  # FIXME
                 [x.price for x in responses if x.side == Side.BUY],
                 '^', color='g')
        ax4.plot([x.time for x in responses if x.side == Side.SELL],  # FIXME
                 [x.price for x in responses if x.side == Side.SELL],
                 'v', color='r')

        # set same limits
        y_bot, y_top = ax1.get_ylim()
        x_bot, x_top = ax1.get_xlim()
        ax1.set_xlim(x_bot, x_top)
        ax2.set_xlim(x_bot, x_top)
        ax3.set_xlim(x_bot, x_top)
        ax4.set_xlim(x_bot, x_top)
        dif = (x_top - x_bot) * .01
        ax1.set_xlim(x_bot - dif, x_top + dif)
        ax2.set_xlim(x_bot - dif, x_top + dif)
        ax3.set_xlim(x_bot - dif, x_top + dif)
        ax4.set_xlim(x_bot - dif, x_top + dif)
        plt.show()



from traitlets import HasTraits, List, Instance, Float, Type, Tuple, Dict, Bool
from .enums import TradingType, ExchangeType, PairType, InstrumentType
from .structs import Instrument


class ExchangeConfig(HasTraits):
    exchange_types = List(trait=Instance(ExchangeType), default_value=[])
    trading_type = Instance(klass=TradingType, args=('NONE',), kwargs={})
    currency_pairs = List(trait=Instance(PairType), default_value=[PairType.BTCUSD])
    instruments = List(trait=Instance(Instrument), default_value=[Instrument(type=InstrumentType.PAIR, underlying=PairType.BTCUSD)])


class SyntheticExchangeConfig(ExchangeConfig):
    exchange_type = Instance(ExchangeType)
    adversaries = List(default_value=[])

    direction = Float(default_value=.5)
    volatility = Float(default_value=200)


class BacktestConfig(HasTraits):
    pass


class RiskConfig(HasTraits):
    max_drawdown = Float(default_value=100.0)  # % Max drawdown before liquidation
    max_risk = Float(default_value=100.0)  # % Max to risk
    total_funds = Float(default_value=0.0)  # total funds available
    trading_type = Instance(klass=TradingType, args=('NONE',), kwargs={})


class ExecutionConfig(HasTraits):
    trading_type = Instance(klass=TradingType, args=('NONE',), kwargs={})


class StrategyConfig(HasTraits):
    clazz = Type()
    args = Tuple(default_value=())
    kwargs = Dict(default_value={})


class TradingEngineConfig(HasTraits):
    type = Instance(klass=TradingType, args=('NONE',), kwargs={})
    print = Bool(default_value=False)
    exchange_options = Instance(klass=ExchangeConfig, args=(), kwargs={})
    backtest_options = Instance(klass=BacktestConfig, args=(), kwargs={})
    risk_options = Instance(klass=RiskConfig, args=(), kwargs={})
    execution_options = Instance(klass=ExecutionConfig, args=(), kwargs={})
    strategy_options = List(trait=Instance(StrategyConfig), default_value=[])  # List of strategy options



from .common import generateTornadoApplication
from ....ui.handlers.trades import TradesHandler
from mock import MagicMock, patch
import tornado.websocket


class TestMessages:
    def setup(self):
        self.app, self.login_code = generateTornadoApplication()

    def test_ServerMessagesMixin(self):
        req = MagicMock()
        req.body = ''

        x = TradesHandler(self.app, req, trading_engine=MagicMock(), psp_kwargs={})
        x._transforms = []

        x.te = MagicMock()
        x.te._ex = MagicMock()
        x.te._ex.messages = MagicMock()
        x.te._ex.messages.return_value = [MagicMock()]
        x.te._ex.messages.return_value[0].to_dict = MagicMock()
        x.te._ex.messages.return_value[0].to_dict.return_value = {'test': 1, 'instrument': {'underlying': 'test'}}

        x.get_data()
        x.get_data('test')
        x.get_data(None, 'BTCUSD')

        x.te._ex.messages.return_value = {'TRADE': [MagicMock()]}
        x.te._ex.messages.return_value['TRADE'][0].to_dict = MagicMock()
        x.te._ex.messages.return_value['TRADE'][0].to_dict.return_value = {'test': 1}

        x.get_data()
        x.get_data(pair='test')
        x.get_data(pair='BTCUSD')

    def test_ServerMessages(self):
        req = MagicMock()
        req.body = ''

        x = TradesHandler(self.app, req, trading_engine=MagicMock(), psp_kwargs={})
        x._transforms = []

        x.te._ex.messages = MagicMock()
        x.te._ex.messages.return_value = [MagicMock()]
        x.te._ex.messages.return_value[0].to_dict = MagicMock()
        x.te._ex.messages.return_value[0].to_dict.return_value = {'test': 1, 'instrument': {'underlying': 'test'}}

        with patch.object(x, 'get_secure_cookie') as mget:
            mget.return_value = self.login_code
            x.get()



import tornado.gen
from .base import HTTPHandler
from tornado.concurrent import run_on_executor
from perspective import PerspectiveHTTPMixin
from aat.enums import PairType
from aat.structs import Instrument


class TradesHandler(PerspectiveHTTPMixin, HTTPHandler):
    '''Server Handler
    Extends:
        HTTPHandler
    '''

    def initialize(self, trading_engine, psp_kwargs=None):
        self.te = trading_engine
        self.psp_kwargs = psp_kwargs or {}

    @run_on_executor
    def get_data(self, exchange=None, pair=None, **psp_kwargs):
        try:
            if pair is not None:
                pair = PairType.from_string(pair)
                instrument = Instrument(underlying=pair)
            else:
                instrument = None
        except (ValueError, TypeError):
            instrument = None

        msgs = [x.to_dict(True, True) for x in self.te.query.query_trades(instrument)]
        if len(msgs) > 0:
            for msg in msgs:
                msg['underlying'] = msg['instrument']['underlying']
                msg['instrument'] = msg['instrument']['underlying'] + ',' + msg['instrument']['type']

        super(TradesHandler, self).loadData(data=msgs, **psp_kwargs)
        return super(TradesHandler, self).getData()

    @tornado.gen.coroutine
    def get(self):
        pair = self.get_argument('pair', '')
        dat = yield self.get_data(pair=pair, **self.psp_kwargs)
        self.write(dat)



from mock import patch, MagicMock


class TestExchange:
    def test_init(self):
        from ...config import ExchangeConfig
        from ...exchanges.gemini import GeminiExchange
        from ...enums import ExchangeType

        with patch('os.environ'), patch('ccxt.gemini') as m:
            ec = ExchangeConfig()
            ec.exchange_type = ExchangeType.GEMINI
            m.get_balances.return_value = []
            e = GeminiExchange(ExchangeType.GEMINI, ec)
            e._running = True
            assert e

    def test_receive(self):
        from ...config import ExchangeConfig
        from ...exchanges.gemini import GeminiExchange
        from ...enums import TickType, ExchangeType

        with patch('os.environ'), patch('ccxt.gemini'):
            ec = ExchangeConfig()
            ec.exchange_type = ExchangeType.GEMINI
            e = GeminiExchange(ExchangeType.GEMINI, ec)
            e._running = True
            assert e

            e.ws = MagicMock()

            with patch('json.loads') as m1:
                for i, val in enumerate([TickType.TRADE,
                                         TickType.OPEN,
                                         TickType.FILL,
                                         TickType.CANCEL,
                                         TickType.CHANGE,
                                         TickType.ERROR]):
                    m1.return_value = {'events': [{"type": "accepted",
                                                   "order_id": "372456298",
                                                   "event_id": "372456299",
                                                   "client_order_id": "20170208_example",
                                                   "api_session": "AeRLptFXoYEqLaNiRwv8",
                                                   "symbol": "btcusd",
                                                   "side": "buy",
                                                   "order_type": "exchange limit",
                                                   "timestamp": "1478203017",
                                                   "timestampms": 1478203017455,
                                                   "is_live": True,
                                                   "is_cancelled": False,
                                                   "is_hidden": False,
                                                   "avg_execution_price": "0",
                                                   "original_amount": "14.0296",
                                                   "price": "1059.54"
                                                   }]}
                    e.receive()



from .trading import TradingEngine
from .parser import parse_command_line_config


def main(argv: list) -> None:
    config = parse_command_line_config(argv)

    # Instantiate trading engine
    #
    # The engine is responsible for managing the different components,
    # including the strategies, the bank/risk engine, and the
    # exchange/backtest engine.

    te = TradingEngine(config)
    # Run the live trading engine
    te.run()
